@@ -89,6 +89,7 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
   def_delegators :@framework_module,
                 :print_status,
                 :print_good,
+                :print_error,
                 :vprint_error,
                 :vprint_status,
                 :workspace
@@ -156,8 +157,9 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
     credential = nil
     if cache_file.present?
       # the cache file is only used for loading credentials, it is *not* written to
-      credential = load_credential_from_file(cache_file, sname: nil, sname_hostname: @hostname)
-      serviceclass = build_spn.name_string.first
+      load_sname_hostname_credential_result = load_credential_from_file(cache_file, sname: nil, sname_hostname: @hostname)
+      credential = load_sname_hostname_credential_result&.fetch(:credential, nil)
+      serviceclass = build_spn&.name_string&.first
       if credential && credential.server.components[0] != serviceclass
         old_sname = credential.server.components.snapshot.join('/')
         credential.server.components[0] = serviceclass
@@ -167,9 +169,20 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
         ticket.sname.name_string[0] = serviceclass
         credential.ticket = ticket.encode
       elsif credential.nil? && hostname.present?
-        credential = load_credential_from_file(cache_file, sname: "krbtgt/#{hostname.split('.', 2).last}")
+        load_sname_krbtgt_hostname_credential_result = load_credential_from_file(cache_file, sname: "krbtgt/#{hostname.split('.', 2).last}")
+        credential = load_sname_krbtgt_hostname_credential_result&.fetch(:credential, nil)
       end
       if credential.nil?
+        print_error("Failed to load a usable credential from ticket file: #{cache_file}")
+        if load_sname_hostname_credential_result
+          print_error("Attempt failed to find a valid credential in #{cache_file} for #{load_sname_hostname_credential_result[:filter].map { |k, v| "#{k}=#{v.inspect}" }.join(', ')}:")
+          print_error(load_sname_hostname_credential_result[:filter_reasons].join("\n").indent(2))
+        end
+
+        if load_sname_krbtgt_hostname_credential_result
+          print_error("Attempt failed to find a valid credential in #{cache_file} for #{load_sname_krbtgt_hostname_credential_result[:filter].map { |k, v| "#{k}=#{v.inspect}" }.join(', ')}")
+          print_error(load_sname_krbtgt_hostname_credential_result[:filter_reasons].join("\n").indent(2))
+        end
         raise ::Rex::Proto::Kerberos::Model::Error::KerberosError.new("Failed to load a usable credential from ticket file: #{cache_file}")
       end
       print_status("Loaded a credential from ticket file: #{cache_file}")
@@ -361,7 +374,7 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
   # @return [Rex::Proto::Kerberos::CredentialCache::Krb5CcacheCredential] The ccache credential
   def request_tgt_only(options = {})
     if options[:cache_file]
-      credential = load_credential_from_file(options[:cache_file])
+      credential = load_credential_from_file(options[:cache_file])&.fetch(:credential, nil)
     else
       credential = get_cached_credential(
         options.merge(
@@ -389,8 +402,8 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
   # @option options [Msf::Exploit::Remote::Kerberos::Ticket::Storage::Base] :ticket_storage Override the @ticket_storage attribute
   # @param [Rex::Proto::Kerberos::CredentialCache::Krb5CcacheCredential] :credential
   #   The ccache credential from the TGT
-  # @see #authenticate_via_krb5_ccache_credential_tgt Options dcoumentation
-  # @see #get_cached_credential Other options dcoumentation
+  # @see #authenticate_via_krb5_ccache_credential_tgt Options documentation
+  # @see #get_cached_credential Other options documentation
   # @return [Rex::Proto::Kerberos::CredentialCache::Krb5CcacheCredential] The ccache credential
   def request_tgs_only(credential, options = {})
     # load a cached TGS
@@ -415,6 +428,7 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
   def s4u2self(credential, options = {})
     realm = self.realm.upcase
     sname = options.fetch(:sname)
+    impersonate_type = options.fetch(:impersonate_Type, 'none')
     client_name = username
 
     now = Time.now.utc
@@ -429,16 +443,22 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
     etypes = Set.new([credential.keyblock.enctype.value])
     etypes << Rex::Proto::Kerberos::Crypto::Encryption::RC4_HMAC
 
+    unless impersonate_type == 'dmsa'
+      pa_data = build_pa_for_user(   {
+                                       username: options[:impersonate],
+                                       session_key: session_key,
+                                       realm: realm
+                                     }
+      )
+    end
+
     tgs_options = {
       ticket_storage: options.fetch(:ticket_storage, @ticket_storage),
       credential_cache_username: options[:impersonate],
-      pa_data: build_pa_for_user(
-        {
-          username: options[:impersonate],
-          session_key: session_key,
-          realm: realm
-        }
-      )
+      pa_data: pa_data,
+      nonce: options[:nonce],
+      impersonate: options[:impersonate],
+      impersonate_type: options[:impersonate_type],
     }
 
     request_service_ticket(
@@ -692,6 +712,9 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
       etype: etypes,
       options: ticket_options,
 
+      impersonate: options[:impersonate],
+      nonce: options[:nonce],
+
       # Specify nil to ensure the KDC uses the current time for the desired starttime of the requested ticket
       from: nil,
       till: expiry_time,
@@ -707,6 +730,9 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
     end
 
     tgs_options = {
+      nonce: options[:nonce],
+      impersonate: options[:impersonate] ,
+      impersonate_type: options[:impersonate_type],
       session_key: session_key,
       subkey: nil,
       checksum: nil,
@@ -718,7 +744,7 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
       body: build_tgs_request_body(**tgs_body_options)
     }
     if options[:pa_data].present?
-      pa_data = [options[:pa_data]] unless options[:pa_data].is_a?(::Enumerable)
+      pa_data = options[:pa_data].is_a?(::Enumerable) ? options[:pa_data] : [options[:pa_data]]
       tgs_options[:pa_data] = pa_data
     end
 
@@ -757,7 +783,7 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
       msg_type: Rex::Proto::Kerberos::Crypto::KeyUsage::TGS_REP_ENCPART_SESSION_KEY
     )
 
-    [tgs_ticket, tgs_auth]
+    [tgs_ticket, tgs_auth, ccache.credentials.first]
   end
 
   private
@@ -828,6 +854,7 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
   def authenticate_via_krb5_ccache_credential_tgt(credential, options = {})
     realm = self.realm.upcase
     sname = options.fetch(:sname)
+    nonce = options.fetch(:nonce)
     client_name = username
 
     now = Time.now.utc
@@ -842,10 +869,11 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
     etypes = Set.new([credential.keyblock.enctype.value])
     tgs_options = {
       pa_data: [],
-      ticket_storage: ticket_storage
+      ticket_storage: ticket_storage,
+      nonce: nonce
     }
 
-    tgs_ticket, tgs_auth = request_service_ticket(
+    tgs_ticket, tgs_auth, credential = request_service_ticket(
       session_key,
       ticket,
       realm,
@@ -866,8 +894,9 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
       do_delegation = tgs_auth.flags.include?(Rex::Proto::Kerberos::Model::KdcOptionFlags::OK_AS_DELEGATE)
     end
 
+    delegated_tgs_ticket = nil
     if do_delegation
-      delegated_tgs_ticket, delegated_tgs_auth = request_delegation_ticket(
+      delegated_tgs_ticket, delegated_tgs_auth, credential = request_delegation_ticket(
         session_key,
         ticket,
         realm,
@@ -911,7 +940,8 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
     {
       service_ap_request: service_ap_request,
       session_key: tgs_auth.key,
-      client_sequence_number: sequence_number
+      client_sequence_number: sequence_number,
+      credential: credential
     }
   end
 
@@ -1035,7 +1065,9 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
       msg_type: Rex::Proto::Kerberos::Crypto::KeyUsage::TGS_REP_ENCPART_SESSION_KEY
     )
 
-    [delegated_tgs_ticket, delegated_tgs_auth]
+    ccache = Rex::Proto::Kerberos::CredentialCache::Krb5Ccache.from_responses(delegated_tgs_res, delegated_tgs_auth)
+
+    [delegated_tgs_ticket, delegated_tgs_auth, ccache.credentials.first]
   end
 
   # Search the database for a credential object that can be used for authentication.
@@ -1054,67 +1086,87 @@ class Msf::Exploit::Remote::Kerberos::ServiceAuthenticator::Base
     )
   end
 
-  # Load a credential object from a file for authentication. Credentials in the file will be filtered by multiple
+  # Load a credential object from a file or database entry for authentication. Credentials in the credential cache will be filtered by multiple
   # attributes including their timestamps to ensure that the returned credential appears usable.
   #
-  # @param [String] file_path The file path to load a credential object from
-  # @return [Rex::Proto::Kerberos::CredentialCache::Krb5CacheCredential] the credential object for authentication
-  def load_credential_from_file(file_path, options = {})
-    unless File.readable?(file_path.to_s)
-      wlog("Failed to load ticket file '#{file_path}' (file not readable)")
-      return nil
-    end
+  # @param [String] path The path to load a credential object from
+  # @return [Hash] :credential [Rex::Proto::Kerberos::CredentialCache::Krb5CacheCredential] the credential object for authentication
+  # @return [Hash] :filter_reasons [Array<String>] the reasons for filtering tickets
+  def load_credential_from_file(path, options = {})
+    # Load a database reference or a path
+    if path&.start_with?('id:')
+      id = path.delete_prefix('id:')
+      storage = Msf::Exploit::Remote::Kerberos::Ticket::Storage::ReadOnly.new(framework: framework)
+      cache = storage.tickets({ id: id  }).first&.ccache
+      unless cache
+        wlog("Invalid cache id #{id} provided")
+        return { credential: nil }
+      end
+    else
+      unless File.readable?(path.to_s)
+        wlog("Failed to load ticket file '#{path}' (file not readable)")
+        return nil
+      end
 
-    begin
-      cache = Rex::Proto::Kerberos::CredentialCache::Krb5Ccache.read(File.binread(file_path))
-    rescue StandardError => e
-      elog("Failed to load ticket file '#{file_path}' (parsing failed)", error: e)
-      return nil
+      begin
+        cache = Rex::Proto::Kerberos::CredentialCache::Krb5Ccache.read(File.binread(path))
+      rescue StandardError => e
+        elog("Failed to load ticket file '#{path}' (parsing failed)", error: e)
+        return nil
+      end
     end
 
     sname = options.fetch(:sname) { build_spn&.to_s }
     sname_hostname = options.fetch(:sname_hostname, nil)
     now = Time.now.utc
 
+    filter = {
+      realm: @realm,
+      sname: sname,
+      sname_hostname: sname_hostname
+    }.merge(options)
+    filter_reasons = []
+
     cache.credentials.to_ary.each.with_index(1) do |credential, index|
       tkt_start = credential.starttime == Time.at(0).utc ? credential.authtime : credential.starttime
       tkt_end = credential.endtime
+      filter_reason_prefix = "Filtered credential #{path} ##{index} reason: "
 
       unless tkt_start < now
-        wlog("Filtered credential #{file_path} ##{index} reason: Ticket start time is before now (start: #{tkt_start})")
+        filter_reasons << "#{filter_reason_prefix}Ticket start time is before now (start: #{tkt_start})"
         next
       end
 
       unless now < tkt_end
-        wlog("Filtered credential #{file_path} ##{index} reason: Ticket is expired (expiration: #{tkt_end})")
+        filter_reasons << "#{filter_reason_prefix}Ticket is expired (expiration: #{tkt_end})"
         next
       end
 
       unless !@realm || @realm.casecmp?(credential.server.realm.to_s)
-        wlog("Filtered credential #{file_path} ##{index} reason: Realm (#{@realm}) does not match (realm: #{credential.server.realm})")
+        filter_reasons << "#{filter_reason_prefix} Realm (#{@realm}) does not match (realm: #{credential.server.realm})"
         next
       end
 
       unless !sname || sname.to_s.casecmp?(credential.server.components.snapshot.join('/'))
-        wlog("Filtered credential #{file_path} ##{index} reason: SPN (#{sname}) does not match (spn: #{credential.server.components.snapshot.join('/')})")
+        filter_reasons << "#{filter_reason_prefix}SPN (#{sname}) does not match (spn: #{credential.server.components.snapshot.join('/')})"
         next
       end
 
       unless !sname_hostname ||
-          sname_hostname.to_s.downcase == credential.server.components[1].downcase ||
-          sname_hostname.to_s.downcase.ends_with?('.' + credential.server.components[1].downcase)
-        wlog("Filtered credential #{file_path} ##{index} reason: SPN (#{sname_hostname}) hostname does not match (spn: #{credential.server.components.snapshot.join('/')})")
+             sname_hostname.to_s.downcase == credential.server.components[1].downcase ||
+             sname_hostname.to_s.downcase.ends_with?('.' + credential.server.components[1].downcase)
+        filter_reasons << "#{filter_reason_prefix}SPN (#{sname_hostname}) hostname does not match (spn: #{credential.server.components.snapshot.join('/')})"
         next
       end
 
       unless !@username || @username.casecmp?(credential.client.components.last.to_s)
-        wlog("Filtered credential #{file_path} ##{index} reason: Username (#{@username}) does not match (username: #{credential.client.components.last})")
+        filter_reasons << "Filtered credential #{path} ##{index} reason: Username (#{@username}) does not match (username: #{credential.client.components.last})"
         next
       end
 
-      return credential
+      return { credential: credential, filter: filter,  filter_reasons: filter_reasons }
     end
 
-    nil
+    { credential: nil, filter: filter, filter_reasons: filter_reasons }
   end
 end
